#!/usr/bin/env python3
from setuptools import setup

setup(name='pyenvisalink',
      version='1.7',
      description='A python3 library for running asynchronus communications with envisalink alarm control panel modules.',
      url='https://github.com/Cinntax/pyenvisalink',
      download_url='https://github.com/Cinntax/pyenvisalink/tarball/1.7',
      author='Dan Cinnamon',
      author_email='dcinnamon@gmail.com',
      license='MIT',
      packages=['pyenvisalink'],
      classifiers=['Development Status :: 4 - Beta', 'Programming Language :: Python :: 3.4','Programming Language :: Python :: 3.5']
      )
